@@ -11,14 +11,21 @@ Spyder kernel for Jupyter.
 """
 
 # Standard library imports
+from distutils.version import LooseVersion
 import os
 import sys
+import threading
 
 # Third-party imports
+import ipykernel
 from ipykernel.ipkernel import IPythonKernel
 
 # Local imports
+from spyder_kernels.py3compat import TEXT_TYPES, to_text_string
 from spyder_kernels.comms.frontendcomm import FrontendComm
+from spyder_kernels.py3compat import PY3, input
+from spyder_kernels.utils.misc import (
+    MPL_BACKENDS_FROM_SPYDER, MPL_BACKENDS_TO_SPYDER, INLINE_FIGURE_FORMATS)
 
 
 # Excluded variables from the Variable Explorer (i.e. they are not
@@ -39,6 +46,7 @@ class SpyderKernel(IPythonKernel):
             'set_breakpoints': self.set_spyder_breakpoints,
             'set_pdb_ignore_lib': self.set_pdb_ignore_lib,
             'set_pdb_execute_events': self.set_pdb_execute_events,
+            'set_pdb_use_exclamation_mark': self.set_pdb_use_exclamation_mark,
             'get_value': self.get_value,
             'load_data': self.load_data,
             'save_namespace': self.save_namespace,
@@ -58,9 +66,19 @@ class SpyderKernel(IPythonKernel):
             'set_namespace_view_settings': self.set_namespace_view_settings,
             'get_var_properties': self.get_var_properties,
             'set_sympy_forecolor': self.set_sympy_forecolor,
-            'set_pdb_echo_code': self.set_pdb_echo_code,
             'update_syspath': self.update_syspath,
-            'is_special_kernel_valid': self.is_special_kernel_valid
+            'is_special_kernel_valid': self.is_special_kernel_valid,
+            'get_matplotlib_backend': self.get_matplotlib_backend,
+            'set_matplotlib_backend': self.set_matplotlib_backend,
+            'set_mpl_inline_figure_format': self.set_mpl_inline_figure_format,
+            'set_mpl_inline_resolution': self.set_mpl_inline_resolution,
+            'set_mpl_inline_figure_size': self.set_mpl_inline_figure_size,
+            'set_mpl_inline_bbox_inches': self.set_mpl_inline_bbox_inches,
+            'set_jedi_completer': self.set_jedi_completer,
+            'set_greedy_completer': self.set_greedy_completer,
+            'set_autocall': self.set_autocall,
+            'pdb_input_reply': self.pdb_input_reply,
+            '_interrupt_eventloop': self._interrupt_eventloop,
             }
         for call_id in handlers:
             self.frontend_comm.register_call_handler(
@@ -70,17 +88,18 @@ class SpyderKernel(IPythonKernel):
 
         self._pdb_obj = None
         self._pdb_step = None
-        self._pdb_print_code = True
         self._do_publish_pdb_state = True
         self._mpl_backend_error = None
         self._running_namespace = None
+        self._pdb_input_line = None
         self.shell.get_local_scope = self.get_local_scope
 
     def get_local_scope(self, stack_depth):
         """Get local scope at given frame depth."""
         frame = sys._getframe(stack_depth + 1)
         if self._pdb_frame is frame:
-            # we also give the globals because they might not be in self.shell.user_ns
+            # we also give the globals because they might not be in
+            # self.shell.user_ns
             namespace = frame.f_globals.copy()
             namespace.update(self._pdb_locals)
             return namespace
@@ -88,7 +107,8 @@ class SpyderKernel(IPythonKernel):
             return frame.f_locals
 
     # -- Public API -----------------------------------------------------------
-    def frontend_call(self, blocking=False, broadcast=True, timeout=None):
+    def frontend_call(self, blocking=False, broadcast=True,
+                      timeout=None, callback=None):
         """Call the frontend."""
         # If not broadcast, send only to the calling comm
         if broadcast:
@@ -99,6 +119,7 @@ class SpyderKernel(IPythonKernel):
         return self.frontend_comm.remote_call(
             blocking=blocking,
             comm_id=comm_id,
+            callback=callback,
             timeout=timeout)
 
     # --- For the Variable Explorer
@@ -266,16 +287,6 @@ class SpyderKernel(IPythonKernel):
             self.frontend_call(blocking=False).pdb_state(state)
         self._do_publish_pdb_state = True
 
-    def pdb_continue(self):
-        """
-        Tell the console to run 'continue' after entering a
-        Pdb session to get to the first breakpoint.
-
-        Fixes issue 2034
-        """
-        if self._pdb_obj:
-            self.frontend_call(blocking=False).pdb_continue()
-
     def set_spyder_breakpoints(self, breakpoints):
         """
         Handle a message from the frontend
@@ -283,14 +294,6 @@ class SpyderKernel(IPythonKernel):
         if self._pdb_obj:
             self._pdb_obj.set_spyder_breakpoints(breakpoints)
 
-    def set_pdb_echo_code(self, state):
-        """Set if pdb should echo the code.
-
-        This might change for each pdb statment and is therefore not included
-        in pdb settings.
-        """
-        self._pdb_print_code = state
-
     def set_pdb_ignore_lib(self, state):
         """
         Change the "Ignore libraries while stepping" debugger setting.
@@ -305,6 +308,66 @@ class SpyderKernel(IPythonKernel):
         if self._pdb_obj:
             self._pdb_obj.pdb_execute_events = state
 
+    def set_pdb_use_exclamation_mark(self, state):
+        """
+        Set an option on the current debugging session to decide wether
+        the Pdb commands needs to be prefixed by '!'
+        """
+        if self._pdb_obj:
+            self._pdb_obj.pdb_use_exclamation_mark = state
+
+    def pdb_input_reply(self, line, echo_stack_entry=True):
+        """Get a pdb command from the frontend."""
+        if self._pdb_obj:
+            self._pdb_obj._disable_next_stack_entry = not echo_stack_entry
+        self._pdb_input_line = line
+        if self.eventloop:
+            # Interrupting the eventloop is only implemented when a message is
+            # received on the shell channel, but this message is queued and
+            # won't be processed because an `execute` message is being
+            # processed. Therefore we process the message here (comm channel)
+            # and request a dummy message to be sent on the shell channel to
+            # stop the eventloop. This will call back `_interrupt_eventloop`.
+            self.frontend_call().request_interrupt_eventloop()
+
+    def cmd_input(self, prompt=''):
+        """
+        Special input function for commands.
+        Runs the eventloop while debugging.
+        """
+        # Only works if the comm is open and this is a pdb prompt.
+        if not self.frontend_comm.is_open() or not self._pdb_frame:
+            return input(prompt)
+
+        # Flush output before making the request.
+        sys.stderr.flush()
+        sys.stdout.flush()
+
+        # Send the input request.
+        self._pdb_input_line = None
+        self.frontend_call().pdb_input(prompt)
+
+        # Allow GUI event loop to update
+        if PY3:
+            is_main_thread = (
+                threading.current_thread() is threading.main_thread())
+        else:
+            is_main_thread = isinstance(
+                threading.current_thread(), threading._MainThread)
+
+        if is_main_thread and self.eventloop:
+            while self._pdb_input_line is None:
+                self.eventloop(self)
+        else:
+            self.frontend_comm.wait_until(
+                lambda: self._pdb_input_line is not None)
+        return self._pdb_input_line
+
+    def _interrupt_eventloop(self):
+        """Interrupts the eventloop."""
+        # Receiving the request is enough to stop the eventloop.
+        pass
+
     # --- For the Help plugin
     def is_defined(self, obj, force_import=False):
         """Return True if object is defined in current namespace"""
@@ -334,6 +397,71 @@ class SpyderKernel(IPythonKernel):
         if valid:
             return getsource(obj)
 
+    # -- For Matplolib
+    def get_matplotlib_backend(self):
+        """Get current matplotlib backend."""
+        try:
+            import matplotlib
+            return MPL_BACKENDS_TO_SPYDER[matplotlib.get_backend()]
+        except Exception:
+            return None
+
+    def set_matplotlib_backend(self, backend, pylab=False):
+        """Set matplotlib backend given a Spyder backend option."""
+        mpl_backend = MPL_BACKENDS_FROM_SPYDER[to_text_string(backend)]
+        self._set_mpl_backend(mpl_backend, pylab=pylab)
+
+    def set_mpl_inline_figure_format(self, figure_format):
+        """Set the inline figure format to use with matplotlib."""
+        mpl_figure_format = INLINE_FIGURE_FORMATS[figure_format]
+        self._set_config_option(
+            'InlineBackend.figure_format', mpl_figure_format)
+
+    def set_mpl_inline_resolution(self, resolution):
+        """Set inline figure resolution."""
+        if LooseVersion(ipykernel.__version__) < LooseVersion('4.5'):
+            option = 'savefig.dpi'
+        else:
+            option = 'figure.dpi'
+        self._set_mpl_inline_rc_config(option, resolution)
+
+    def set_mpl_inline_figure_size(self, width, height):
+        """Set inline figure size."""
+        value = (width, height)
+        self._set_mpl_inline_rc_config('figure.figsize', value)
+
+    def set_mpl_inline_bbox_inches(self, bbox_inches):
+        """
+        Set inline print figure bbox inches.
+
+        The change is done by updating the ´rint_figure_kwargs' config dict.
+        """
+        from IPython.core.getipython import get_ipython
+        config = get_ipython().kernel.config
+        inline_config = (
+            config['InlineBackend'] if 'InlineBackend' in config else {})
+        print_figure_kwargs = (
+            inline_config['print_figure_kwargs']
+            if 'print_figure_kwargs' in inline_config else {})
+        bbox_inches_dict = {
+            'bbox_inches': 'tight' if bbox_inches else None}
+        print_figure_kwargs.update(bbox_inches_dict)
+        self._set_config_option(
+            'InlineBackend.print_figure_kwargs', print_figure_kwargs)
+
+    # -- For completions
+    def set_jedi_completer(self, use_jedi):
+        """Enable/Disable jedi as the completer for the kernel."""
+        self._set_config_option('IPCompleter.use_jedi', use_jedi)
+
+    def set_greedy_completer(self, use_greedy):
+        """Enable/Disable greedy completer for the kernel."""
+        self._set_config_option('IPCompleter.greedy', use_greedy)
+
+    def set_autocall(self, autocall):
+        """Enable/Disable autocall funtionality."""
+        self._set_config_option('ZMQInteractiveShell.autocall', autocall)
+
     # --- Additional methods
     def set_cwd(self, dirname):
         """Set current working directory."""
@@ -562,6 +690,8 @@ class SpyderKernel(IPythonKernel):
 
         backend: A parameter that can be passed to %matplotlib
                  (e.g. 'inline' or 'tk').
+        pylab: Is the pylab magic should be used in order to populate the
+               namespace from numpy and matplotlib
         """
         import traceback
         from IPython.core.getipython import get_ipython
@@ -604,6 +734,37 @@ class SpyderKernel(IPythonKernel):
 
         self._mpl_backend_error = error
 
+    def _set_config_option(self, option, value):
+        """
+        Set config options using the %config magic.
+
+        As parameters:
+            option: config option, for example 'InlineBackend.figure_format'.
+            value: value of the option, for example 'SVG', 'Retina', etc.
+        """
+        from IPython.core.getipython import get_ipython
+        try:
+            base_config = "{option} = "
+            value_line = (
+                "'{value}'" if isinstance(value, TEXT_TYPES) else "{value}")
+            config_line = base_config + value_line
+            get_ipython().run_line_magic(
+                'config',
+                config_line.format(option=option, value=value))
+        except Exception:
+            pass
+
+    def _set_mpl_inline_rc_config(self, option, value):
+        """
+        Update any of the Matplolib rcParams given an option and value.
+        """
+        try:
+            from matplotlib import rcParams
+            rcParams[option] = value
+        except Exception:
+            # Needed in case matplolib isn't installed
+            pass
+
     def show_mpl_backend_errors(self):
         """Show Matplotlib backend errors after the prompt is ready."""
         if self._mpl_backend_error is not None:
